// Basic utility functions.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# define MAXLINEWIDTH 100000
# define ENTRYWIDTH 50
# define PHYLIPWIDTH 10
# define SPECIESNAMELENGTH 50
# define max(i,j) (i>j ? i : j)
# define argmax(i,j) (j>i ? 2 : 1)

int inlist(int Nv, int *v, int target);
char * getitemval2(char *s, int ind, int *length);
char * getitemval3(char *s, int ind, int *length, char sepch);
double atof2(char *s, int length);
int * setdiff(int NA, int *A, int NB, int *B, int *NC);
int nu2ind(char ch);
int aa2ind(char ch);

extern int nchar, njointchar;
extern char *numap, *aamap;

// Check if a number is on a list.
int inlist(int Nv, int *v, int target)
{
  int i=0, indicator=0;

  while ((i<Nv)&&(indicator==0))
    {
      if (*(v+i)==target)
	indicator=1;
      i++;
    }

  return indicator;
}

// Get the designated item in a string.
// Difference from getitemval: items can be separated by tab or space.
char * getitemval2(char *s, int ind, int *length)
{
  int i, tabcnt, curind, cnt;
  char ch, *item, *template;

  template=malloc(sizeof(char)*MAXLINEWIDTH); tabcnt=0; curind=0;

  while ((tabcnt<ind)&&(curind<strlen(s)))
    {
      ch=*(s+curind);

      if ((ch==0x09)||(ch==' '))
	tabcnt++;
      // If the line ends before tabcnt reaching ind, then return null.
      else if ((ch==0x0a)&&(tabcnt<ind))
	tabcnt=ind+1;
      curind++;
    }

  if (tabcnt==ind)
    {      
      cnt=0;
      while (((ch=*(s+curind))!=0x09)&&(ch!=0x0a)&&(ch!=0x0d)&&(ch!=' '))
	{
	  *(template+cnt)=ch;
	  curind++; cnt++;
	}
    
      if (cnt>0)
	{
	  //item=malloc(sizeof(char)*cnt);
	  item=malloc(sizeof(char)*(cnt+1));
	  for (i=0; i<=(cnt-1); i++)
	    *(item+i)=*(template+i);

	  *(item+cnt)='\0';

	  *length=cnt;
	}
      else
	*length=cnt;
           
    }
  else
    {
      item=NULL; *length=0;
    }

  free(template);
  return item;
}

// Get the designated item in a string.
// Difference from getitemval2: specify the separator character.
char * getitemval3(char *s, int ind, int *length, char sepch)
{
  int i, tabcnt, curind, cnt;
  char ch, *item, *template;

  template=malloc(sizeof(char)*MAXLINEWIDTH); tabcnt=0; curind=0;

  while ((tabcnt<ind)&&(curind<strlen(s)))
    {
      ch=*(s+curind);

      if (ch==sepch)
	tabcnt++;
      // If the line ends before tabcnt reaching ind, then return null.
      else if ((ch==0x0a)&&(tabcnt<ind))
	tabcnt=ind+1;
      curind++;
    }

  if (tabcnt==ind)
    {
      
      cnt=0;
      while (((ch=*(s+curind))!=sepch)&&(ch!=0x0a)&&(ch!=0x0d))
	{
	  *(template+cnt)=ch;
	  curind++; cnt++;
	}
    
      if (cnt>0)
	{
	  item=malloc(sizeof(char)*cnt);
	  for (i=0; i<=(cnt-1); i++)
	    *(item+i)=*(template+i);
	  *length=cnt;
	}
      else
	*length=cnt;
           
    }
  else
    {
      item=NULL; *length=0;
    }

  free(template);
  
  return item;
}

// Convert a string into a real number.
// Also works for exponential representation.
double atof2(char *s, int length)
{
  double val, expval, power;
  int i, j, cnt, sign, afterpoint, afterexp, expsign;
  char ch;

  val=0; sign=1; cnt=0; afterpoint=0; afterexp=0; 
  expval=0, expsign=1; power=1.0;
  //while (cnt<strlen(s))
  while (cnt<length)
    {
      ch=*(s+cnt);
      if ((cnt==0)&&(ch=='-'))
	sign=-1;
      else if ((ch=='-')&&(afterexp==1))
	expsign=-1;
      else if (ch=='.')
	{
	  afterpoint=1;
	  power=0.1;
	}
      else if ((ch=='e')||(ch=='E'))
	{	
	  afterexp=1;
	  afterpoint=0;
	}
      else if ((ch>='0')&&(ch<='9')&&(afterpoint==0))
	{
	  if (afterexp==0)
	    val=val*10+(ch-'0');
	  else
	    expval=expval*10+(ch-'0');
	}
      else if ((ch>='0')&&(ch<='9')&&(afterpoint==1))
	{
	  if (afterexp==0)
	    val=val+(ch-'0')*power;
	  else
	    expval=expval+(ch-'0')*power;
	  power=power*0.1;
	}
  
      cnt++;
    }

  
  if (sign==-1)
    val=val*(-1);
  if (expsign==-1)
    expval=expval*(-1);

  val=val*pow(10,expval);

  return val;
}

// Obtain a set difference A\B.
int * setdiff(int NA, int *A, int NB, int *B, int *NC)
{
  int i, j, k, *C;
 
  *NC=0; C=malloc(sizeof(int)*max(NA,NB));
  for (i=0; i<=(NA-1); i++)
    {
      j=0; k=1;
      while ((j<=(NB-1))&&(k==1))
	{
	  if (*(A+i)==*(B+j))
	    k=0;
	  j++;
	}
      if (k==1)
	{
	  *NC+=1; *(C+*NC-1)=*(A+i);
	}
    }

  return C;
}

// Convert a nucleic acid character into an index number.
int nu2ind(char ch)
{
  int ind;
  switch (ch) {
    case 'A': ind=1; break;
    case 'C': ind=2; break;
    case 'G': ind=3; break;
    case 'U': ind=4; break;
    default: ind=0; break;
  }
  return ind;
}

// Convert an amino acid character into an index number.
int aa2ind(char ch)
{
  int ind;
  switch (ch) {
    case 'A': ind=1; break;
    case 'R': ind=2; break;
    case 'N': ind=3; break;
    case 'D': ind=4; break;
    case 'C': ind=5; break;
    case 'Q': ind=6; break;
    case 'E': ind=7; break;
    case 'G': ind=8; break;
    case 'H': ind=9; break;
    case 'I': ind=10; break;
    case 'L': ind=11; break;
    case 'K': ind=12; break;
    case 'M': ind=13; break;
    case 'F': ind=14; break;
    case 'P': ind=15; break;
    case 'S': ind=16; break;
    case 'T': ind=17; break;
    case 'W': ind=18; break;
    case 'Y': ind=19; break;
    case 'V': ind=20; break;
    default: ind=0; break;
  }

  return ind;
}
