/* Extract domain sequences and species for the test inputs of the downloadable program.
   Need to be compiled in UCSC servers.
 */

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "db.h"
# define MAXLINEWIDTH 100000
# define NAMELENGTH 30
# define LONGNAMELENGTH 100
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define argmax(i,j) (j>i ? 2 : 1)

extern char * getitemval2(char *s, int ind, int *length);
extern char *i2a(long num, int *length);
extern double atof2(char *s, int length);
extern int aa2ind(char ch);
extern char * getitemval3(char *s, int ind, int *length, char sepch);
extern void loadindextree(char *indexfile, int *Nnodes, int *Nproteins, int *Nleaves, int *index_tree_pnt, int *leaves_pnt);
extern void loadindextree2(char *indexfile, int *Nnodes, int *Nproteins, int *Nleaves, int *index_tree_pnt, int *leaves_pnt);
extern int inlist(int Nv, int *v, int target);
extern void get_sequence_tree(char *pfamacc1, char *pfamacc2, FILE *fpfamaccs, FILE **fpfam, int Npfam_nodes, int Npfam_proteins, int Npfam_leaves, struct index_node *pfam_index_tree, int *pfam_leaves, FILE *fsprotaccs, FILE *fsprot, int Nsprot_proteins, int Nsprot_nodes, int Nsprot_leaves, struct index_node *sprot_index_tree, int *sprot_leaves, FILE *ftremblaccs, int Ntremblfiles, FILE **ftrembl, int Ntrembl_proteins, int Ntrembl_nodes, int Ntrembl_leaves, struct index_node *trembl_index_tree, int *trembl_leaves, FILE *fsortedlineage, FILE *flineage, int Nlineage_species, int Nlineage_nodes, int Nlineage_leaves, struct index_node *lineage_index_tree, int *lineage_leaves, char *treedir, int *retnseq, int *retlseq1, int *retlseq2, int *domains1_pnt, int *ranges1_pnt, int *accs1_pnt, int *seqs1_pnt, int *domains2_pnt, int *ranges2_pnt, int *accs2_pnt, int *seqs2_pnt, int *species_pnt, int *retnnodes, int *child_pnt, int *t_pnt, int *order_pnt);
extern void calculate_joint_rate(double *q1, double *q2, double *q2d);
extern double * kron(int nrows1, int ncols1, double *mat1, int nrows2, int ncols2, double *mat2);
extern void padm(int ndim, double *A, double t, int p, double *E);
extern double llratio_quantized_rates2(int nnodes, int **child, double *t1, double *t2, int equalrate, int *order, int nseq, char *seqs, int Nbllevels, double *bllevels, double *prior2, double *nulltransitions, double *activetransitions);
extern void getproteinposition_fileid2(char *acc, FILE *fp, int Nnodes, int Nproteins, int Nleaves, struct index_node *index_tree, int *leaves, int acclength, int shift, int *fileid, long *position);
extern void getpfamentry6(FILE *fpfam, long fileposition, int *nseq, int *lseq, int *domainnames_pnt, int *accs_pnt, int *ranges_pnt, int *seqs_pnt);
void getspeciesname(FILE *fsprotaccs, FILE *fsprot, int Nsprot_proteins, int Nsprot_nodes, int Nsprot_leaves, struct index_node *sprot_index_tree,int *sprot_leaves, FILE *ftremblaccs, int Ntremblfiles, FILE **ftrembl, int Ntrembl_proteins, int Ntrembl_nodes, int Ntrembl_leaves, struct index_node *trembl_index_tree, int *trembl_leaves, int nseq, char *accs, int *retnspecis, int *species_pnt, int *domain2species_pnt);
extern void getgenetree(char *rawdomains, char *rawaccs, int *rawranges, int nseq, int nsubset, int *subset, char *treedir, char *pfamacc, int *retnnodes, int *child_pnt, int *t_pnt);


void recurse_generate_reg_string2(int nnodes, int **child, char *names, int namelength, double *t, int curnode, int *stringlength, char *treestring);

int nseq=0, lseq=0, nnodes=0, nchar=20, njointchar=400, gshift;
char *aamap=" ARNDCQEGHILKMFPSTWYV";
//double epsilon=0.1, timescale=0.01, logrthre=1.0;
double epsilon=0.7, timescale=0.01, logrthre=1.0;
int accstringlength=6;
int shift, acclength;
int globalpos1, globalpos2;

int main(int argc, char *argv[])
{
  // *(argv+1) -- pfam index file.
  // *(argv+2) -- pfam sorted accs and positions.
  // *(argv+3) -- prefix names of separated pfam database files.
  // *(argv+4) -- number of separated pfam files.  
  // *(argv+5) -- swissprot index file.
  // *(argv+6) -- swissprot sorted accs and positions.
  // *(argv+7) -- swissprot database file.
  // *(argv+8) -- trembl index file.
  // *(argv+9) -- trembl sorted accs, file ids and positions.
  // *(argv+10) -- prefix names of separated trembl database files.
  // *(argv+11) -- the number of separated trembl database files.
  // *(argv+12) -- directory of pfam family trees.
  // *(argv+13) -- taxonomy index file for each species of lineage information.
  // *(argv+14) -- species names and positions sorted by alphabets.
  // *(argv+15) -- lineage information file.
  // *(argv+16) -- filtered pfam family pairs.
  // *(argv+17) -- the designated pfam pair.
  // *(argv+18) -- output 1: names and sequences of domain family 1.
  // *(argv+19) -- output 2: names and sequences of domain family 2.
  // *(argv+20) -- output 3: gene tree 1.
  // *(argv+21) -- output 4: gene tree 2.
  // *(argv+22) -- output 5: species names.
  // *(argv+23) -- output 6: species tree.
  // *(argv+24) -- output 7: domain 1 to species mapping.
  // *(argv+25) -- output 8: domain 2 to species mapping.

  char ch, *s, *item, *prefix, *filename, *treedir, *pfampairs;
  int i, j, k, l, m, n, Npfamfiles, Ntremblfiles, length, flag;
  int pnt1, pnt2, pnt3, pnt4, pnt5, pnt6, pnt7, pnt8, pnt9, pnt10, pnt11, pnt12;
  int Nsprot_proteins, Nsprot_nodes, Nsprot_leaves, *sprot_leaves;
  int Ntrembl_proteins, Ntrembl_nodes, Ntrembl_leaves, *trembl_leaves;
  int Npfam_proteins, Npfam_nodes, Npfam_leaves, *pfam_leaves;
  int Nlineage_species, Nlineage_nodes, Nlineage_leaves, *lineage_leaves;  
  struct index_node *pfam_index_tree, *sprot_index_tree, *trembl_index_tree, *lineage_index_tree;
  FILE *fpfamaccs, **fpfam, *fsprotaccs, *fsprot, *ftremblaccs, **ftrembl, *fsortedlineage, *flineage, *fout, *fp, *fp2;  
  double val, tmpval, singleconsthre, consthre, covarthre, gapthre, scorethre;  
  double *prior, *prior2, *q1, *q2, *q2d, *nulltransitions, *activetransitions;  
  int Nbllevels, startline, nlines, endline, nvalids;
  double *bllevels, qval;
  time_t tp1, tp2, *tp;
  int nseq1, nseq2, lseq1, lseq2, nspecies, *gene2sp1, *gene2sp2, ngenetreenodes1, **genetreechild1, ngenetreenodes2, **genetreechild2, nsptreenodes, **speciestreechild;
  char *famnames1, *famnames2, *species;
  double *t1, *t2, *t;
  char *domains1, *domains2, *accs1, *accs2, *seqs1, *seqs2, *species1, *species2, *domain2species1, *domain2species2;
  int *ranges1, *ranges2, *order, nsubset1, *subset1, nsubset2, *subset2;
  char *rawdomains1, *rawaccs1, *rawseqs1, *rawdomains2, *rawaccs2, *rawseqs2, *pfamacc1, *pfamacc2, *subaccs1, *subaccs2, *treestring;
  int *rawranges1, *rawranges2, nspecies1, nspecies2, stringlength, nspeciestreenodes;
  long fileid, curpos;

  // Load the pfam index tree.
  loadindextree(*(argv+1),&Npfam_nodes,&Npfam_proteins,&Npfam_leaves,&pnt1,&pnt2);
  pfam_index_tree=pnt1; pfam_leaves=pnt2;
  
  // Load the pfam sorted accs and positions.
  if ((fpfamaccs=fopen(*(argv+2),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+2));
      return -1;
    }
  
  // Load pfam database files.
  prefix=malloc(sizeof(char)*(strlen(*(argv+3))+1)); 
  strncpy(prefix,*(argv+3),strlen(*(argv+3)));
  *(prefix+strlen(*(argv+3)))='\0';
  Npfamfiles=atoi(*(argv+4));
  
  fpfam=malloc(sizeof(FILE)*Npfamfiles); filename=malloc(sizeof(char));
  for (i=0; i<=(Npfamfiles-1); i++)
    {      
      item=i2a(i,&length);
      filename=realloc(filename,sizeof(char)*(strlen(prefix)+length+1));
      strncpy(filename,prefix,strlen(prefix));
      strncpy(filename+strlen(prefix),item,length);
      *(filename+strlen(prefix)+length)='\0';
      free(item);      
      if ((*(fpfam+i)=fopen(filename,"r"))==NULL)
	{
	  printf("Cannot open %s.\n",filename);
	  return -1;
	}                
    }
  free(filename);    
      
  // Load the swissprot index tree.
  loadindextree(*(argv+5),&Nsprot_nodes,&Nsprot_proteins,&Nsprot_leaves,&pnt1,&pnt2);
  sprot_index_tree=pnt1; sprot_leaves=pnt2;
    
  // Load the swissprot sorted accs and positions.
  if ((fsprotaccs=fopen(*(argv+6),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+6));
      return -1;
    }

  // Load the swissprot database file.
  if ((fsprot=fopen(*(argv+7),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+7));
      return -1;
    }

  // Load the trembl index tree.
  loadindextree(*(argv+8),&Ntrembl_nodes,&Ntrembl_proteins,&Ntrembl_leaves,&pnt1,&pnt2);
  trembl_index_tree=pnt1; trembl_leaves=pnt2;
  
  // Load the trembl sorted accs, file ids and positions.
  if ((ftremblaccs=fopen(*(argv+9),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+9));
      return -1;
    }
  
  // Load the trembl database files.
  prefix=malloc(sizeof(char)*(strlen(*(argv+10))+1));
  strncpy(prefix,*(argv+10),strlen(*(argv+10)));
  *(prefix+strlen(*(argv+10)))='\0';
  Ntremblfiles=atoi(*(argv+11));
  ftrembl=malloc(sizeof(int)*Ntremblfiles); filename=malloc(sizeof(char));
  for (i=0; i<=(Ntremblfiles-1); i++)
    {
      item=i2a(i,&length);
      filename=realloc(filename,sizeof(char)*(strlen(prefix)+length+1));
      strncpy(filename,prefix,strlen(prefix));
      strncpy(filename+strlen(prefix),item,length);
      *(filename+strlen(prefix)+length)='\0';
      free(item);            
      if ((*(ftrembl+i)=fopen(filename,"r"))==NULL)
	{
	  printf("Cannot open %s.\n",filename);
	  return -1;
	}
    }
  free(filename);
  
  // Load the tree directory.
  treedir=malloc(sizeof(char)*(1+strlen(*(argv+12))));
  strncpy(treedir,*(argv+12),strlen(*(argv+12)));
  *(treedir+strlen(*(argv+12)))='\0';

  // Load the lineage index tree.
  loadindextree2(*(argv+13),&Nlineage_nodes,&Nlineage_species,&Nlineage_leaves,&pnt1,&pnt2);
  lineage_index_tree=pnt1; lineage_leaves=pnt2;
  
  // Load the sorted lineage species.
  if ((fsortedlineage=fopen(*(argv+14),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+14));
      return -1;
    }
  
  // Load the lineage information file.
  if ((flineage=fopen(*(argv+15),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+15));
      return -1;
    }
  
  /*
  // Load the pfam pairs.
  pfampairs=malloc(sizeof(char)*7*2);
  if ((fp=fopen(*(argv+16),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+16));
      return -1;
    }
  s=malloc(sizeof(char)*MAXLINEWIDTH);  

  startline=atoi(*(argv+17)); 

  n=0;
  while ((fgets(s,MAXLINEWIDTH,fp)!=NULL)&&(n<startline))
    n++;
  fgets(s,MAXLINEWIDTH,fp);
  
  // Generate the joint tree and extract sequences together.
  item=getitemval2(s,0,&length);
  strncpy(pfampairs,item,7);
  free(item);
  item=getitemval2(s,1,&length);
  strncpy(pfampairs+7,item,7);
  free(item);
  */

  pfampairs=malloc(sizeof(char)*14);
  strncpy(pfampairs,*(argv+17),14);

  // Debug
  printf("pfampairs=%s\n",pfampairs);
  

  get_sequence_tree(pfampairs,pfampairs+7,fpfamaccs,fpfam,Npfam_nodes,Npfam_proteins,Npfam_leaves,pfam_index_tree,pfam_leaves,fsprotaccs,fsprot,Nsprot_proteins,Nsprot_nodes,Nsprot_leaves,sprot_index_tree,sprot_leaves,ftremblaccs,Ntremblfiles,ftrembl,Ntrembl_proteins,Ntrembl_nodes,Ntrembl_leaves,trembl_index_tree,trembl_leaves,fsortedlineage,flineage,Nlineage_species,Nlineage_nodes,Nlineage_leaves,lineage_index_tree,lineage_leaves,treedir,&nseq,&lseq1,&lseq2,&pnt1,&pnt2,&pnt3,&pnt4,&pnt5,&pnt6,&pnt7,&pnt8,&pnt9,&nspeciestreenodes,&pnt10,&pnt11,&pnt12);
  domains1=pnt1; ranges1=pnt2; accs1=pnt3; seqs1=pnt4;
  domains2=pnt5; ranges2=pnt6; accs2=pnt7; seqs2=pnt8;
  species=pnt9; speciestreechild=pnt10; t=pnt11; order=pnt12;

  // Debug
  printf("nseq=%d, lseq1=%d, lseq2=%d, nspeciestreenodes=%d.\n",nseq,lseq1,lseq2,nspeciestreenodes);
  

  // Find the domain family members which belong to designated species.
  pfamacc1=pfampairs; pfamacc2=pfampairs+7;
  
  // Get the start position of the pfam entry 1.
  shift=3; acclength=7;
  getproteinposition_fileid2(pfamacc1,fpfamaccs,Npfam_nodes,Npfam_proteins,Npfam_leaves,pfam_index_tree,pfam_leaves,acclength,shift,&fileid,&curpos);
  
  // Get the aligned sequences and protein names and ranges of pfam family 1.
  if ((fileid>=0)&&(curpos>=0))
    {
      getpfamentry6(*(fpfam+fileid),curpos,&nseq1,&lseq1,&pnt1,&pnt2,&pnt3,&pnt4);
      rawdomains1=pnt1; rawaccs1=pnt2; rawranges1=pnt3; rawseqs1=pnt4;
    }  

  // Get the start position of the pfam entry 2.
  shift=3; acclength=7;
  getproteinposition_fileid2(pfamacc2,fpfamaccs,Npfam_nodes,Npfam_proteins,Npfam_leaves,pfam_index_tree,pfam_leaves,acclength,shift,&fileid,&curpos);

  // Get the aligned sequences and protein names and ranges of pfam family 2.
  if ((fileid>=0)&&(curpos>=0))
    {
      getpfamentry6(*(fpfam+fileid),curpos,&nseq2,&lseq2,&pnt1,&pnt2,&pnt3,&pnt4);
      rawdomains2=pnt1; rawaccs2=pnt2; rawranges2=pnt3; rawseqs2=pnt4;
    }
  
  // Debug
  printf("nseq1=%d, lseq1=%d, nseq2=%d, lseq2=%d\n",nseq1,lseq1,nseq2,lseq2);
  
  
  // Get the species name of each domain member.
  getspeciesname(fsprotaccs,fsprot,Nsprot_proteins,Nsprot_nodes,Nsprot_leaves,sprot_index_tree,sprot_leaves,ftremblaccs,Ntremblfiles,ftrembl,Ntrembl_proteins,Ntrembl_nodes,Ntrembl_leaves,trembl_index_tree,trembl_leaves,nseq1,rawaccs1,&nspecies1,&pnt1,&pnt2);
  species1=pnt1; domain2species1=pnt2;        
  getspeciesname(fsprotaccs,fsprot,Nsprot_proteins,Nsprot_nodes,Nsprot_leaves,sprot_index_tree,sprot_leaves,ftremblaccs,Ntremblfiles,ftrembl,Ntrembl_proteins,Ntrembl_nodes,Ntrembl_leaves,trembl_index_tree,trembl_leaves,nseq2,rawaccs2,&nspecies2,&pnt1,&pnt2);
  species2=pnt1; domain2species2=pnt2;

  // Debug
  printf("nspecies1=%d, nspecies2=%d\n",nspecies1,nspecies2);

  
  // Acquire a subset from each domain family which are contained in the designated species.
  // Avoid choosing the domains of the same proteins.
  nsubset1=0; subset1=malloc(sizeof(int)); gene2sp1=malloc(sizeof(int));
  for (i=0; i<=(nseq1-1); i++)
    {
      j=0; k=-1;
      while ((j<nseq)&&(k==-1))
	{
	  if (strncmp(domain2species1+i*LONGNAMELENGTH,species+j*LONGNAMELENGTH,max(strlen(domain2species1+i*LONGNAMELENGTH),strlen(species+j*LONGNAMELENGTH)))==0)
	    k=j;
	  j++;
	}
      
      // Check if the domain already exists.
      if (k>=0)
	{
	  l=0; m=-1;
	  while ((l<nsubset1)&&(m==-1))
	    {
	      if (strncmp(rawaccs1+i*NAMELENGTH,rawaccs1+*(subset1+l)*NAMELENGTH,6)==0)
		m=l;
	      l++;
	    }
	  if (m>=0)
	    k=-1;
	}

      if (k>=0)
	{
	  nsubset1++; subset1=realloc(subset1,sizeof(int)*nsubset1);
	  *(subset1+nsubset1-1)=i;
	  gene2sp1=realloc(gene2sp1,sizeof(int)*nsubset1);
	  *(gene2sp1+nsubset1-1)=k;
	}
    }

  nsubset2=0; subset2=malloc(sizeof(int)); gene2sp2=malloc(sizeof(int));
  for (i=0; i<=(nseq2-1); i++)
    {
      j=0; k=-1;
      while ((j<nseq)&&(k==-1))
	{
	  if (strncmp(domain2species2+i*LONGNAMELENGTH,species+j*LONGNAMELENGTH,max(strlen(domain2species2+i*LONGNAMELENGTH),strlen(species+j*LONGNAMELENGTH)))==0)
	    k=j;
	  j++;
	}

      // Check if the domain already exists.
      if (k>=0)
	{
	  l=0; m=-1;
	  while ((l<nsubset2)&&(m==-1))
	    {
	      if (strncmp(rawaccs2+i*NAMELENGTH,rawaccs2+*(subset2+l)*NAMELENGTH,6)==0)
		m=l;
	      l++;
	    }
	  if (m>=0)
	    k=-1;
	}

      if (k>=0)
	{
	  nsubset2++; subset2=realloc(subset2,sizeof(int)*nsubset2);
	  *(subset2+nsubset2-1)=i;
	  gene2sp2=realloc(gene2sp2,sizeof(int)*nsubset2);
	  *(gene2sp2+nsubset2-1)=k;
	}
    }

  // Debug
  printf("nsubset1=%d, nsubset2=%d\n",nsubset1,nsubset2);
    
  // Get the gene trees of the two domain families.
  // Indices of the leaf nodes are the indices of domains in subset1 and subset2.
  getgenetree(rawdomains1,rawaccs1,rawranges1,nseq1,nsubset1,subset1,treedir,pfamacc1,&ngenetreenodes1,&pnt1,&pnt2);
  genetreechild1=pnt1; t1=pnt2;
  getgenetree(rawdomains2,rawaccs2,rawranges2,nseq2,nsubset2,subset2,treedir,pfamacc2,&ngenetreenodes2,&pnt1,&pnt2);
  genetreechild2=pnt1; t2=pnt2;
  
  // Debug
  printf("ngenetreenodes1=%d, ngenetreenodes2=%d\n",ngenetreenodes1,ngenetreenodes2);
   
  // Report selected names and sequences of family 1.
  if ((fout=fopen(*(argv+18),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+18));
      return -1;
    }  
  fprintf(fout,"%d %d\n",nsubset1,lseq1);
  for (i=0; i<=(nsubset1-1); i++)
    {
      for (j=0; j<=(6-1); j++)
	fprintf(fout,"%c",*(rawaccs1+*(subset1+i)*NAMELENGTH+j));
      fprintf(fout,"  ");
      for (j=0; j<=(lseq1-1); j++)
	fprintf(fout,"%c",*(rawseqs1+*(subset1+i)*lseq1+j));
      fprintf(fout,"\n");
    }
  fclose(fout);

  // Report selected names and sequences of family 2.
  if ((fout=fopen(*(argv+19),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+19));
      return -1;
    }  
  fprintf(fout,"%d %d\n",nsubset2,lseq2);
  for (i=0; i<=(nsubset2-1); i++)
    {
      for (j=0; j<=(6-1); j++)
	fprintf(fout,"%c",*(rawaccs2+*(subset2+i)*NAMELENGTH+j));
      fprintf(fout,"  ");
      for (j=0; j<=(lseq2-1); j++)
	fprintf(fout,"%c",*(rawseqs2+*(subset2+i)*lseq2+j));
      fprintf(fout,"\n");
    }
  fclose(fout);  
  
  // Report gene tree 1.  
  // Generate the tree string.
  subaccs1=malloc(sizeof(char)*nsubset1*NAMELENGTH);
  memset(subaccs1,' ',nsubset1*NAMELENGTH);  
  for (i=0; i<=(nsubset1-1); i++)
    {
      strncpy(subaccs1+i*NAMELENGTH,rawaccs1+*(subset1+i)*NAMELENGTH,NAMELENGTH);
      *(subaccs1+i*NAMELENGTH+6)='\0';
    }
  
  treestring=malloc(sizeof(char)*MAXLINEWIDTH*10); stringlength=0;
  memset(treestring,'_',MAXLINEWIDTH*10);
  
  recurse_generate_reg_string2(ngenetreenodes1,genetreechild1,subaccs1,NAMELENGTH,t1,ngenetreenodes1-1,&stringlength,treestring);
  *(treestring+stringlength)='\0';
  
  if ((fout=fopen(*(argv+20),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+20));
      return -1;
    } 
  fprintf(fout,"%s\n",treestring);
  fclose(fout);

  // Report gene tree 2.  
  // Generate the tree string.
  subaccs2=malloc(sizeof(char)*nsubset2*NAMELENGTH);
  memset(subaccs2,' ',nsubset2*NAMELENGTH);
  for (i=0; i<=(nsubset2-1); i++)
    {
      strncpy(subaccs2+i*NAMELENGTH,rawaccs2+*(subset2+i)*NAMELENGTH,NAMELENGTH);
      *(subaccs2+i*NAMELENGTH+6)='\0';
    }

  //treestring=malloc(sizeof(char)*MAXLINEWIDTH*10); 
  stringlength=0;
  memset(treestring,'_',MAXLINEWIDTH*10);
  recurse_generate_reg_string2(ngenetreenodes2,genetreechild2,subaccs2,NAMELENGTH,t2,ngenetreenodes2-1,&stringlength,treestring);
  *(treestring+stringlength)='\0';
  
  if ((fout=fopen(*(argv+21),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+21));
      return -1;
    } 
  fprintf(fout,"%s\n",treestring);
  fclose(fout);
      
  // Report selected species names.
  if ((fout=fopen(*(argv+22),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+22));
      return -1;
    } 
  for (i=0; i<=(nseq-1); i++)
    fprintf(fout,"%s\n",species+i*LONGNAMELENGTH);
  fclose(fout);  

  // Generate and report the species tree.
  for (i=0; i<=(nspeciestreenodes-1); i++)
    *(t+i)=0;
  treestring=malloc(sizeof(char)*MAXLINEWIDTH*10); stringlength=0;
  memset(treestring,'_',MAXLINEWIDTH*10);  
  
  for (i=0; i<=(nseq-1); i++)
    {
      for (j=0; j<strlen(species+i*LONGNAMELENGTH); j++)
	if (*(species+i*LONGNAMELENGTH+j)==' ')
	  *(species+i*LONGNAMELENGTH+j)='_';
    }  

  recurse_generate_reg_string2(nspeciestreenodes,speciestreechild,species,LONGNAMELENGTH,t,nspeciestreenodes-1,&stringlength,treestring);
  *(treestring+stringlength)='\0';

  if ((fout=fopen(*(argv+23),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+23));
      return -1;
    } 
  fprintf(fout,"%s\n",treestring);
  fclose(fout);

  for (i=0; i<=(nseq-1); i++)
    {
      for (j=0; j<strlen(species+i*LONGNAMELENGTH); j++)
	if (*(species+i*LONGNAMELENGTH+j)=='_')
	  *(species+i*LONGNAMELENGTH+j)=' ';
    }  

  // Report the mapping from gene 1 to species.
  if ((fout=fopen(*(argv+24),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+24));
      return -1;
    } 
  for (i=0; i<=(nsubset1-1); i++)    
    fprintf(fout,"%s%c%s\n",subaccs1+i*NAMELENGTH,0x09,species+*(gene2sp1+i)*LONGNAMELENGTH);
  fclose(fout);

  // Report the mapping from gene 2 to species.
  if ((fout=fopen(*(argv+25),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+25));
      return -1;
    } 
  for (i=0; i<=(nsubset2-1); i++)    
    fprintf(fout,"%s%c%s\n",subaccs2+i*NAMELENGTH,0x09,species+*(gene2sp2+i)*LONGNAMELENGTH);
  fclose(fout);

  return 0;
}



// Recursively generate a regular expression string of a tree.
// Difference from recurse_generate_reg_string: inputs are integers rather than the phylogenetic_node structure.
void recurse_generate_reg_string2(int nnodes, int **child, char *names, int namelength, double *t, int curnode, int *stringlength, char *treestring)
{
  int i, j, k, l;
  char ch;
    
  // An internal node.
  if (**(child+curnode)>0)  
    {
      // Put a left parenthesis before the children.
      *stringlength+=1; *(treestring+*stringlength-1)='(';
      
      for (i=1; i<=**(child+curnode); i++)
	{
	  // Recurse to each child.
	  recurse_generate_reg_string2(nnodes,child,names,namelength,t,*(*(child+curnode)+i),stringlength,treestring);
	  
	  // Separate children by commas.
	  if (i<**(child+curnode))	  
	    {
	      *stringlength+=1; *(treestring+*stringlength-1)=',';
	    }
	}
     
      //  Put a right parenthesis after the children.
      *stringlength+=1; *(treestring+*stringlength-1)=')';
        
      // If not a root then put the branch length.
      if (curnode<(nnodes-1))
	{
	  char *s;
	  *stringlength+=1; *(treestring+*stringlength-1)=':';
	  s=malloc(sizeof(char)*MAXLINEWIDTH);
	  sprintf(s,"%.5f",*(t+curnode));	  
	  strncpy(treestring+*stringlength,s,7);
	  *stringlength+=7;
	  free(s);
	}

      // If a root then put a semicolumn at the end.
      else
	{
	  *stringlength+=1; *(treestring+*stringlength-1)=';';
	}      
    }

  // A leaf node.
  else
    {
      char *s;

      // Put the species name.
      /*
      k=0;
      while ((k<=(NAMELENGTH-1))&&((ch=*((nodes+curnode)->name->s+k))!=' '))
	k++;
      strncpy(treestring+*stringlength,(nodes+curnode)->name->s,k);
      *stringlength+=k;
      */

      // Debug
      k=strlen(names+curnode*namelength);
      strncpy(treestring+*stringlength,names+curnode*namelength,k);
      *stringlength+=k;      
      
      // Put the branch length.
      *stringlength+=1; *(treestring+*stringlength-1)=':';
      s=malloc(sizeof(char)*MAXLINEWIDTH);
      sprintf(s,"%.5f",*(t+curnode));
      strncpy(treestring+*stringlength,s,7);
      *stringlength+=7;
      free(s);    
    }
}      
	  
