// Utility functions involved in phylogeny.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "phylogeny.h"
# define MAXLINEWIDTH 100000
# define NAMELENGTH 100
# define LONGNAMELENGTH 100
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define argmax(i,j) (j>i ? 2 : 1)

extern int inlist(int Nv, int *v, int target);
extern int * setdiff(int NA, int *A, int NB, int *B, int *NC);
extern double atof2(char *s, int length);

void parsetree(char *treestring, int nseq, char *names, int *child_pnt, int *t_pnt, int *order_pnt);
void get_sequence_tree(int nseq1, int lseq1, char *accs1, char *seqs1, int *gene2sp1, int ngenetreenodes1, int **genetreechild1, double *t1, int nseq2, int lseq2, char *accs2, char *seqs2, int *gene2sp2, int ngenetreenodes2, int **genetreechild2, double *t2, int nspecies, char *speciesnames, int nsptreenodes, int **sptreechild, int *retnseq, int *subaccs1_pnt, int *subseqs1_pnt, int *subaccs2_pnt, int *subseqs2_pnt, int *t_pnt);
void reconciliation(int nspecies, int nspeciestreenodes, int **speciestreechild, int ngenes, int *gene2species, int ngenetreenodes, int **genetreechild, int *mapping, int *duplabel);
void select_representatives(int ngenes, int nspecies, int *gene2species, int ngenetreenodes, int **genetreechild, int *duplabel, int *retnrepresentatives, int *representatives_pnt);
int recurse_max_coverage(int curnode, int nspecies, int *gene2species, int **genetreechild, int *duplabel, int **descendants, int *selection, int *representatives_pnt);
void get_joint_tree_bl(int nspecies, int nspeciestreenodes, int **speciestreechild, int *gene2species1, int ngenetreenodes1, int **genetreechild1, double *t1, int *mapping1, int *gene2species2, int ngenetreenodes2, int **genetreechild2, double *t2, int *mapping2, int nrep, int *rep1, int *rep2, int *jointt_pnt);
void gene2speciesbl(int ngenetreenodes, int **genetreechild, double *genet, int nrep, int *rep, int nspeciestreenodes, int nspecies, int **speciestreechild, int *mapping, double *speciest);
void recurse_dist(int ngenetreenodes, int **genetreechild, int *mapping, int curnode, int s, double t, double *speciest);
void traverseorder(int nnodes, int **child, int *order_pnt);
void partialtraverseorder(int nnodes, int **child, int nrep, int *rep, int *retn, int *order_pnt);

extern int nchar, njointchar, nnodes;
extern double epsilon;

void parsetree(char *treestring, int nseq, char *names, int *child_pnt, int *t_pnt, int *order_pnt)
{
  int i, j, k, nss, **child, *order, length, curnode, readbl, *perm, node, npushin, npopup;
  double *t, val;
  char *temp, ch;
  struct cellstr *ss;
  struct phylogenetic_node *nodes;
      
  // First round: separate species, distances and characters
  nss=0; ss=malloc(sizeof(struct cellstr)); length=0;
  temp=malloc(sizeof(char)*MAXLINEWIDTH);
   
  for (i=0; i<=(strlen(treestring)-1); i++)
    {      
      ch=*(treestring+i);
           
      // Delimiter.
      if ((ch=='(')||(ch==')')||(ch==':')||(ch==' ')||(ch==',')||(ch==';'))
	{	  
	  if (length>0)
	    {	      
	      nss++; 
	      ss=realloc(ss,sizeof(struct cellstr)*nss);	      
	      (ss+nss-1)->length=length;	      
	      (ss+nss-1)->s=malloc(sizeof(char)*length);	      
	      strncpy((ss+nss-1)->s,temp,length);	      
	      length=0;
	      	      
	      if ((ch!=' ')&&(ch!=';'))
		{
		  nss++;
		  ss=realloc(ss,sizeof(struct cellstr)*nss);
		  (ss+nss-1)->length=1;
		  (ss+nss-1)->s=malloc(sizeof(char));
		  strncpy((ss+nss-1)->s,treestring+i,1);
		}	      
	    }	  
	  else
	    {	      
	      if ((ch!=' ')&&(ch!=';'))
		{
		  nss++;
		  ss=realloc(ss,sizeof(struct cellstr)*nss);
		  (ss+nss-1)->length=1;
		  (ss+nss-1)->s=malloc(sizeof(char));
		  strncpy((ss+nss-1)->s,treestring+i,1);
		}	      
	    }	  
	}
        
      // Content.
      else
	{
	  length++; *(temp+length-1)=ch;	 
	}
      
    }

  free(temp);
    
  // Check if the treestring is valid.
  npushin=0; npopup=0;
  for (i=0; i<=(nss-1); i++)
    {
      // left parenthesis.
      if (((ss+i)->length==1)&&(*((ss+i)->s)=='('))
	npushin++;
      // right parenthesis: return to the parent node.
      else if (((ss+i)->length==1)&&(*((ss+i)->s)==')'))
	npopup++;
      // comma: retrun to the parent node.
      else if (((ss+i)->length==1)&&(*((ss+i)->s)==','))
	npopup++;
      // text: create a new leaf node and put the species name.
      else if (((*((ss+i)->s)>='A')&&(*((ss+i)->s)<='Z'))||((*((ss+i)->s)>='a')&&(*((ss+i)->s)<='z')))
	npushin++;
    }

  if (npushin!=(npopup+1))
    {
      for (i=0; i<=(nss-1); i++)
	free((ss+i)->s);
      free(ss);
      nnodes=0; *child_pnt=0; *t_pnt=0; *order_pnt=0;
      return;
    }  
  
  // Second run: recursively build the tree.
  curnode=-1; nnodes=0; nodes=malloc(sizeof(struct phylogenetic_node));
  readbl=0;
 
  for (i=0; i<=(nss-1); i++)  
    {
      
      // left parenthesis: create an internal node and attach it as a child under the curnode.  Set curnode as the new node.  If curnode=0 then create a root node.
      if (((ss+i)->length==1)&&(*((ss+i)->s)=='('))
	{
	  nnodes++;
	  nodes=realloc(nodes,sizeof(struct phylogenetic_node)*nnodes);
	  
	  (nodes+nnodes-1)->ind=nnodes-1; 
	  (nodes+nnodes-1)->parent=curnode;
	  (nodes+nnodes-1)->nchild=0; 
	  (nodes+nnodes-1)->bl=0;
	  
	  if (curnode>=0)
	    {
	      (nodes+curnode)->nchild+=1;
	      if ((nodes+curnode)->nchild==1)
		(nodes+curnode)->child=malloc(sizeof(int));
	      else
		(nodes+curnode)->child=realloc((nodes+curnode)->child,sizeof(int)*(nodes+curnode)->nchild);
	      *((nodes+curnode)->child+(nodes+curnode)->nchild-1)=nnodes-1;
	    }	

	  curnode=nnodes-1;	    	  
	}

      // right parenthesis: return to the parent node.
      else if (((ss+i)->length==1)&&(*((ss+i)->s)==')'))
	{
	  curnode=(nodes+curnode)->parent;	  
	}

      // comma: retrun to the parent node.
      else if (((ss+i)->length==1)&&(*((ss+i)->s)==','))
	{	  
	  curnode=(nodes+curnode)->parent;
	}

      // text: create a new leaf node and put the species name.
      else if (((*((ss+i)->s)>='A')&&(*((ss+i)->s)<='Z'))||((*((ss+i)->s)>='a')&&(*((ss+i)->s)<='z')))
	{
	  nnodes++;
	  nodes=realloc(nodes,sizeof(struct phylogenetic_node)*nnodes);
	  (nodes+nnodes-1)->ind=nnodes-1; 
	  (nodes+nnodes-1)->parent=curnode;
	  (nodes+nnodes-1)->nchild=0; 
	  (nodes+nnodes-1)->bl=0;
	  (nodes+nnodes-1)->name=malloc(sizeof(struct cellstr));
	  (nodes+nnodes-1)->name->length=(ss+i)->length;
	  (nodes+nnodes-1)->name->s=malloc(sizeof(char)*(ss+i)->length);
	  strncpy((nodes+nnodes-1)->name->s,(ss+i)->s,(ss+i)->length);	  

	  // Replace '_' with ' ' in species names.
	  for (k=0; k<(ss+i)->length; k++)
	    if (*((nodes+nnodes-1)->name->s+k)=='_')
	      *((nodes+nnodes-1)->name->s+k)=' '; 

	  if (curnode>=0)
	    {
	      (nodes+curnode)->nchild+=1;
	      if ((nodes+curnode)->nchild==1)
		(nodes+curnode)->child=malloc(sizeof(int));
	      else
		(nodes+curnode)->child=realloc((nodes+curnode)->child,sizeof(int)*(nodes+curnode)->nchild);
	      *((nodes+curnode)->child+(nodes+curnode)->nchild-1)=nnodes-1;
	    }	  
	  curnode=nnodes-1;	  
	}
      
      // column: start to read the branch length of the current node.
      else if (((ss+i)->length==1)&&(*((ss+i)->s)==':'))
	{
	  readbl=1;	  
	}

      // number after column: branch length.
      else if (readbl==1)
	{
	  readbl=0;
	  val=atof2((ss+i)->s,(ss+i)->length);
	  (nodes+curnode)->bl=val;
	}      
    }

  // Permute node indices such that the indices of leaves conform with the order in species names.
  // Also the last node is the root.
  perm=malloc(sizeof(int)*nnodes);
  for (i=0; i<=(nnodes-1); i++)
    *(perm+i)=-1;

  // Put the leaf nodes and root in the right positions.
  for (i=0; i<=(nseq-1); i++)
    {
      int ind1=-1, ind2=-1;

      j=0;
      while ((j<=(nnodes-1))&&(ind1==-1))
	{
	  if (((nodes+j)->nchild==0)&&(strncmp(names+i*NAMELENGTH,(nodes+j)->name->s,max(strlen(names+i*NAMELENGTH),(nodes+j)->name->length))==0))
	    ind1=j;
	  j++;
	}                            
      *(perm+ind1)=i;      
    }
  *(perm+0)=nnodes-1;

  // Place the remaining nodes.
  {
    int *temp1, *temp2, *temp3, *temp4;

    temp1=malloc(sizeof(int)*nnodes);
    temp2=malloc(sizeof(int)*nnodes);
    
    for (i=0; i<=(nnodes-1); i++)
      {
	*(temp1+i)=-1; *(temp2+i)=-1;
      }
    for (i=0; i<=(nnodes-1); i++)
      {
	if (*(perm+i)>=0)
	  {
	    *(temp1+i)=1; *(temp2+*(perm+i))=1;
	  }
      }       

    temp3=malloc(sizeof(int)*(nnodes-1-nseq));
    temp4=malloc(sizeof(int)*(nnodes-1-nseq));         
    
    j=0; k=0; 
    for (i=0; i<=(nnodes-1); i++)
      {
	if (*(temp1+i)==-1)
	  {
	    j++; *(temp3+j-1)=i;	    
	  }	
	if (*(temp2+i)==-1)
	  {
	    k++; *(temp4+k-1)=i;	   	    
	  }	
      }

    for (i=0; i<=(nnodes-2-nseq); i++)
      *(perm+*(temp3+i))=*(temp4+i);
  
    free(temp1); free(temp2); free(temp3); free(temp4);    
  }

  for (i=0; i<=(nnodes-1); i++)
    {
      (nodes+i)->ind=*(perm+(nodes+i)->ind);
      if ((nodes+i)->parent>=0)
	(nodes+i)->parent=*(perm+(nodes+i)->parent);
      for (j=0; j<=((nodes+i)->nchild-1); j++)
	*((nodes+i)->child+j)=*(perm+*((nodes+i)->child+j));
    }

  // Obtain the tree topology.
  child=malloc(sizeof(int)*nnodes);
  for (i=0; i<=(nnodes-1); i++)
    {
      j=(nodes+i)->ind;
      *(child+j)=malloc(sizeof(int)*(1+(nodes+i)->nchild));
      **(child+j)=(nodes+i)->nchild;
      
      for (k=1; k<=(nodes+i)->nchild; k++)
	*(*(child+j)+k)=*((nodes+i)->child+k-1);
    }

  // Obtain the branch length.
  t=malloc(sizeof(double)*nnodes);
  for (i=0; i<=(nnodes-1); i++)
    {
      j=(nodes+i)->ind;
      *(t+j)=(nodes+i)->bl;
    }

  // Obtain the traversing order of the tree.
  order=malloc(sizeof(int)*nnodes); node=nnodes-1; length=0;
  while (length<nnodes)
    {
      int *childinds, *remain;
      
      curnode=-1; i=0;
      while ((i<=(nnodes-1))&&(curnode==-1))
	{
	  if ((nodes+i)->ind==node)
	    curnode=i;
	  i++;
	}      

      childinds=malloc(sizeof(int)*((nodes+curnode)->nchild+1));
      for (i=0; i<=((nodes+curnode)->nchild-1); i++)
	*(childinds+i)=*((nodes+curnode)->child+i);
      
      remain=setdiff((nodes+curnode)->nchild,childinds,length,order,&k);
      
      if (((nodes+curnode)->nchild==0)||(k==0))
	{
	  length++; *(order+length-1)=(nodes+curnode)->ind;
	  if (node!=(nnodes-1))
	    node=(nodes+curnode)->parent;
	}
      else
	node=*remain;
      free(childinds); free(remain);
    }

  // Release memory.  
  for (i=0; i<=(nss-1); i++)
    free((ss+i)->s);
  free(ss);  
  for (i=0; i<=(nnodes-1); i++)
    {
      if ((nodes+i)->nchild==0)
	{
	  free((nodes+i)->name->s); free((nodes+i)->name);
	}
      else if ((nodes+i)->nchild>0)
	free((nodes+i)->child);
    }
  free(nodes);
  free(perm);
    
  // Return the values.
  *child_pnt=child; *t_pnt=t; *order_pnt=order;  

  return;
}

// Extract representative genes and their sequences of two gene families.
// Also obtain the joint phylogenetic tree using reconciliation.
// Assume all species in the list appear in both families.  So the species tree does not change.
void get_sequence_tree(int nseq1, int lseq1, char *accs1, char *seqs1, int *gene2sp1, int ngenetreenodes1, int **genetreechild1, double *t1, int nseq2, int lseq2, char *accs2, char *seqs2, int *gene2sp2, int ngenetreenodes2, int **genetreechild2, double *t2, int nspecies, char *speciesnames, int nsptreenodes, int **sptreechild, int *retnseq, int *subaccs1_pnt, int *subseqs1_pnt, int *subaccs2_pnt, int *subseqs2_pnt, int *t_pnt)
{
  int i, j, k, l, m, n;
  int *mapping1, *mapping2, *duplabel1, *duplabel2;
  int *rep1, *rep2, pnt1, pnt2;
  char *subaccs1, *subaccs2, *subseqs1, *subseqs2;
  double *t;

  // Build reconciliation of the two gene trees.
  mapping1=malloc(sizeof(int)*ngenetreenodes1); 
  duplabel1=malloc(sizeof(int)*ngenetreenodes1);
  reconciliation(nspecies,nsptreenodes,sptreechild,nseq1,gene2sp1,ngenetreenodes1,genetreechild1,mapping1,duplabel1);

  mapping2=malloc(sizeof(int)*ngenetreenodes2); 
  duplabel2=malloc(sizeof(int)*ngenetreenodes2);
  reconciliation(nspecies,nsptreenodes,sptreechild,nseq2,gene2sp2,ngenetreenodes2,genetreechild2,mapping2,duplabel2);
  
  // Select representative domains from the reconciliation trees.
  // Guarantees to select nspecies representatives from each family.
  select_representatives(nseq1,nspecies,gene2sp1,ngenetreenodes1,genetreechild1,duplabel1,&k,&pnt1);
  rep1=pnt1;
  select_representatives(nseq2,nspecies,gene2sp2,ngenetreenodes2,genetreechild2,duplabel2,&k,&pnt1);
  rep2=pnt1;  
 
  // Write the selected gene names and sequences.
  subaccs1=malloc(sizeof(char)*nspecies*NAMELENGTH);
  subseqs1=malloc(sizeof(char)*nspecies*lseq1);
  for (i=0; i<=(nspecies-1); i++)
    {
      strncpy(subaccs1+i*NAMELENGTH,accs1+*(rep1+i)*NAMELENGTH,NAMELENGTH);
      strncpy(subseqs1+i*lseq1,seqs1+*(rep1+i)*lseq1,lseq1);
    }
  subaccs2=malloc(sizeof(char)*nspecies*NAMELENGTH);
  subseqs2=malloc(sizeof(char)*nspecies*lseq2);
  for (i=0; i<=(nspecies-1); i++)
    {
      strncpy(subaccs2+i*NAMELENGTH,accs2+*(rep2+i)*NAMELENGTH,NAMELENGTH);
      strncpy(subseqs2+i*lseq2,seqs2+*(rep2+i)*lseq2,lseq2);
    }
  
  // Calculate the branch length of the joint tree.
  get_joint_tree_bl(nspecies,nsptreenodes,sptreechild,gene2sp1,ngenetreenodes1,genetreechild1,t1,mapping1,gene2sp2,ngenetreenodes2,genetreechild2,t2,mapping2,nspecies,rep1,rep2,&pnt1);
  t=pnt1; 

  // Release memory.
  free(mapping1); free(mapping2);
  free(duplabel1); free(duplabel2);
  free(rep1); free(rep2);

  // Return.
  *retnseq=nspecies; *subaccs1_pnt=subaccs1; *subseqs1_pnt=subseqs1;
  *subaccs2_pnt=subaccs2; *subseqs2_pnt=subseqs2; *t_pnt=t;
  
  return;
}

// Build reconciliation between a gene tree and a species tree.
// Mark the nodes in the gene tree where duplication occurs.
// Notice both the gene tree and species tree are binary.
void reconciliation(int nspecies, int nspeciestreenodes, int **speciestreechild, int ngenes, int *gene2species, int ngenetreenodes, int **genetreechild, int *mapping, int *duplabel)
{
  int i, j, k, l, *order, *speciesparents, pnt1, *descendants;

  // Get the parents of nodes in species tree.
  speciesparents=malloc(sizeof(int)*nspeciestreenodes);
  for (i=0; i<=(nspeciestreenodes-1); i++)
    *(speciesparents+i)=-1;
  for (i=0; i<=(nspeciestreenodes-1); i++)
    {
      for (j=1; j<=**(speciestreechild+i); j++)
	{
	  k=*(*(speciestreechild+i)+j);
	  *(speciesparents+k)=i;
	}
    }

  // Obtain the traverse order of the gene tree from leaves to the root.
  traverseorder(ngenetreenodes,genetreechild,&pnt1);
  order=pnt1;

  // Mark the gene-species node mapping of the leaves of the gene tree.
  for (i=0; i<=(ngenetreenodes-1); i++)
    *(mapping+i)=-1;
  for (i=0; i<=(ngenes-1); i++)
    *(mapping+i)=*(gene2species+i);

  // Initialize the duplication labels.
  for (i=0; i<=(ngenetreenodes-1); i++)
    *(duplabel+i)=0;

  // Initialize the descendant information.
  descendants=malloc(sizeof(int)*ngenetreenodes*nspecies);
  for (i=0; i<=(ngenetreenodes*nspecies-1); i++)
    *(descendants+i)=0;
  for (i=0; i<=(ngenes-1); i++)
    *(descendants+i*nspecies+*(gene2species+i))=1;

  // Recursively mark the gene-species node mapping and duplication labeles.
  for (i=0; i<=(ngenetreenodes-1); i++)
    {
      int g=*(order+i);

      // If g is an internal node.
      if (**(genetreechild+g)>0)
	{
	  int g1, g2, a, b;

	  // Get the two children of g.
	  g1=*(*(genetreechild+g)+1); g2=*(*(genetreechild+g)+2);
	  
	  // Get the gene-species tree mapping of the two children.
	  a=*(mapping+g1); b=*(mapping+g2);
	  
	  // Find the gene-species tree mapping of g.
	  // Traverse a and b to their common ancestor.
	  while (a!=b)
	    {
	      if (a<b)
		a=*(speciesparents+a);
	      else
		b=*(speciesparents+b);
	    }
	  *(mapping+g)=a;

	  // When g and its child map to the same species node.
	  if ((*(mapping+g)==*(mapping+g1))||(*(mapping+g)==*(mapping+g2)))
	    {
	      // Scenario 1: the two children of g have overlapped species.  Label it as a duplication node.
	      k=0;
	      for (j=0; j<=(nspecies-1); j++)
		k+=(*(descendants+g1*nspecies+j)*(*(descendants+g2*nspecies+j)));
	      if (k>0)
		*(duplabel+g)=1;

	      // Scenario 2: the two children of g do not have overlapped species.  Label it as a inconsistent speciation node.
	      else
		*(duplabel+g)=2;
	    }
	  
	  // Update the descendant information.
	  // The descendants of g are the union of descendants of g1 and g2.
	  for (j=0; j<=(nspecies-1); j++)
	    *(descendants+g*nspecies+j)=((*(descendants+g1*nspecies+j)==1)||(*(descendants+g2*nspecies+j)==1));	  
	}
    }	  
  
  // Release memory.
  free(speciesparents); free(order); free(descendants);

  return;
}

// From a gene tree, a species tree and their reconciliation, find the representative genes which belong to the same duplication lineage and cover the largest number of species.
// Apply a dynamic programming to select representatives.  At a speciation node, the score is the sum of the scores of its two children.  At a duplication node, the score is the supremum of the scores of its two children.
// At a duplication node, allows to select genes under both children if the current node is not the duplication node for those genes.
void select_representatives(int ngenes, int nspecies, int *gene2species, int ngenetreenodes, int **genetreechild, int *duplabel, int *retnrepresentatives, int *representatives_pnt)
{
  int i, j, k, nrepresentatives, *representatives, *selection, pnt1, **descendants, *order;

  // Obtain the descendants of each node.
  traverseorder(ngenetreenodes,genetreechild,&pnt1);
  order=pnt1;

  descendants=malloc(sizeof(int)*ngenetreenodes);  
  for (i=0; i<=(ngenes-1); i++)
    {
      *(descendants+i)=malloc(sizeof(int)*2); **(descendants+i)=1;
      *(*(descendants+i)+1)=i;
    }
  for (i=0; i<=(ngenetreenodes-1); i++)
    {
      int g=*(order+i);
     
      if (**(genetreechild+g)==2)
	{
	  int g1, g2;
	  g1=*(*(genetreechild+g)+1); g2=*(*(genetreechild+g)+2);
	  *(descendants+g)=malloc(sizeof(int)*(1+**(descendants+g1)+**(descendants+g2)));
	  **(descendants+g)=**(descendants+g1)+**(descendants+g2);
	  for (j=1; j<=**(descendants+g1); j++)
	    *(*(descendants+g)+j)=*(*(descendants+g1)+j);
	  for (j=1; j<=**(descendants+g2); j++)
	    *(*(descendants+g)+j+**(descendants+g1))=*(*(descendants+g2)+j);
	}
    }
  free(order);

  // Initialize the selection vector.  All genes are selected.
  selection=malloc(sizeof(int)*ngenes);
  for (i=0; i<=(ngenes-1); i++)
    *(selection+i)=1;

  // Apply a dynamic programming algorithm to find the maximum coverage representatives and to label the selection vector.
  nrepresentatives=recurse_max_coverage(ngenetreenodes-1,nspecies,gene2species,genetreechild,duplabel,descendants,selection,&pnt1);
  representatives=pnt1;
  
  // Report the selected representatives.  
  free(representatives);
    
  // The selected representatives follow the order of species.
  nrepresentatives=nspecies; representatives=malloc(sizeof(int)*nspecies);
  for (i=0; i<=(nspecies-1); i++)
    *(representatives+i)=-1;
  for (i=0; i<=(ngenes-1); i++)
    {
      if (*(selection+i)==1)
	{
	  j=*(gene2species+i);
	  *(representatives+j)=i;
	}
    }

  /*
  nrepresentatives=0; representatives=malloc(sizeof(int)*ngenes);
  for (i=0; i<=(ngenes-1); i++)
    {
      if (*(selection+i)==1)
	{
	  nrepresentatives++; *(representatives+nrepresentatives-1)=i;
	}
    }
  representatives=realloc(representatives,sizeof(int)*nrepresentatives);
  */

  free(selection);
  for (i=0; i<=(ngenetreenodes-1); i++)
    free(*(descendants+i));
  free(descendants);
  *retnrepresentatives=nrepresentatives; *representatives_pnt=representatives;
  return;
}

// Recursively obtain the maximum coverage representatives and label the selection vector.
// Notice here the representatives are used to choose duplication branches rather than determining the final representatives.
int recurse_max_coverage(int curnode, int nspecies, int *gene2species, int **genetreechild, int *duplabel, int **descendants, int *selection, int *representatives_pnt)
{
  int i, j, k, nrep, nrep1, nrep2, *rep1, *rep2, *representatives, pnt1;  

  // A leaf node: return the gene as the representative.
  // The gene is not deselected.
  if (**(genetreechild+curnode)==0)
    {
      nrep=1; representatives=malloc(sizeof(int));
      *representatives=curnode; *representatives_pnt=representatives;
      return nrep;     
    }

  // An internal node.    
  else if (**(genetreechild+curnode)==2)
    {
      int g1, g2;

      // Get the coverage and representatives of the two children.
      g1=*(*(genetreechild+curnode)+1);
      nrep1=recurse_max_coverage(g1,nspecies,gene2species,genetreechild,duplabel,descendants,selection,&pnt1);
      rep1=pnt1;
      g2=*(*(genetreechild+curnode)+2);
      nrep2=recurse_max_coverage(g2,nspecies,gene2species,genetreechild,duplabel,descendants,selection,&pnt1);
      rep2=pnt1;
      
      // Both representatives are null.
      if ((nrep1<=0)&&(nrep2<=0))
	nrep=0;

      // A speciation node: add the coverage score and put the two representative sets together.
      // No gene is deselected.
      else if ((*(duplabel+curnode)==0)||(*(duplabel+curnode)==2))
	{	  	  
	  nrep=nrep1+nrep2;
	  representatives=malloc(sizeof(int)*nrep);
	  for (i=0; i<=(nrep1-1); i++)
	    *(representatives+i)=*(rep1+i);
	  for (i=0; i<=(nrep2-1); i++)
	    *(representatives+nrep1+i)=*(rep2+i);	  
	}

      // A duplication node: select the child with better coverage.
      // For the genes which are duplicated at the current node, deselect the genes under the minor child.
      else if (*(duplabel+curnode)==1)
	{
	  int *speciesmask;
	  	  
	  // Find the species which are duplicated at the current node.
	  speciesmask=malloc(sizeof(int)*nspecies*2);
	  for (i=0; i<=(nspecies*2-1); i++)
	    *(speciesmask+i)=0;
	  for (i=1; i<=**(descendants+g1); i++)
	    *(speciesmask+*(gene2species+*(*(descendants+g1)+i)))=1;
	  for (i=1; i<=**(descendants+g2); i++)
	    *(speciesmask+nspecies+*(gene2species+*(*(descendants+g2)+i)))=1;
	  

	  // Choose the duplication representatives and deselect the duplicated genes under the minor child.
	  if (nrep1>=nrep2)
	    {	      
	      nrep=nrep1;
	      representatives=malloc(sizeof(int)*nrep);
	      for (i=0; i<=(nrep1-1); i++)
		*(representatives+i)=*(rep1+i);

	      for (i=1; i<=**(descendants+g2); i++)
		{
		  k=*(*(descendants+g2)+i);
		  if (*(speciesmask+*(gene2species+k))==1)
		    *(selection+k)=0;
		}	      
	    }
	  else
	    {	      
	      nrep=nrep2;
	      representatives=malloc(sizeof(int)*nrep);
	      for (i=0; i<=(nrep2-1); i++)
		*(representatives+i)=*(rep2+i);

	      for (i=1; i<=**(descendants+g1); i++)
		{
		  k=*(*(descendants+g1)+i);
		  if (*(speciesmask+nspecies+*(gene2species+k))==1)
		    *(selection+k)=0;
		}
	    }
	  free(speciesmask);
	}

      // Erroroneous case.
      else
	nrep=0;

      // Release memory.
      if (nrep1>0)
	free(rep1); 
      if (nrep2>0)
	free(rep2);

      // Return.
      *representatives_pnt=representatives;
      return nrep;
    }

  // Erroroneous case.
  else
    {
      nrep=0; return nrep;
    }
}

// Calculate the branch length of the joint tree.
void get_joint_tree_bl(int nspecies, int nspeciestreenodes, int **speciestreechild, int *gene2species1, int ngenetreenodes1, int **genetreechild1, double *t1, int *mapping1, int *gene2species2, int ngenetreenodes2, int **genetreechild2, double *t2, int *mapping2, int nrep, int *rep1, int *rep2, int *jointt_pnt)
{
  int i, j, k, l, n, njointnodes, **jointchild;
  int *repspecies, pnt1, pnt2, pnt3;
  double *speciest1, *speciest2, *speciest, *jointt;

  // Obtain the branch length of the two gene trees in the selected representatives mapping to the species tree.
  speciest1=malloc(sizeof(double)*nspeciestreenodes);
  speciest2=malloc(sizeof(double)*nspeciestreenodes);
  speciest=malloc(sizeof(double)*nspeciestreenodes);
  for (i=0; i<=(nspeciestreenodes-1); i++)
    {
      *(speciest1+i)=0; *(speciest2+i)=0;
    }
  gene2speciesbl(ngenetreenodes1,genetreechild1,t1,nspecies,rep1,nspeciestreenodes,nspecies,speciestreechild,mapping1,speciest1);
  gene2speciesbl(ngenetreenodes2,genetreechild2,t2,nspecies,rep2,nspeciestreenodes,nspecies,speciestreechild,mapping2,speciest2);
  
  // Average the branch length of the two families.
  for (i=0; i<=(nspeciestreenodes-1); i++)
    *(speciest+i)=0.5*(*(speciest1+i)+*(speciest2+i));

  // Release memory.
  free(speciest1); free(speciest2); 

  *jointt_pnt=speciest;
  return;
}

// Map the branch length in a gene tree to the branch length in a species tree.
void gene2speciesbl(int ngenetreenodes, int **genetreechild, double *genet, int nrep, int *rep, int nspeciestreenodes, int nspecies, int **speciestreechild, int *mapping, double *speciest)
{
  int i, j, k, l, m, n, *parents, *order, pnt1, ngenes, *sparents;

  // Obtain the partial traverse order on the representative leaves.
  partialtraverseorder(ngenetreenodes,genetreechild,nrep,rep,&n,&pnt1);
  order=pnt1;
  
  // Get the parents of the species tree.
  sparents=malloc(sizeof(int)*nspeciestreenodes);
  for (i=0; i<=(nspeciestreenodes-1); i++)
    *(sparents+i)=-1;
  for (i=0; i<=(nspeciestreenodes-1); i++)
    {
      for (j=1; j<=**(speciestreechild+i); j++)
	{
	  k=*(*(speciestreechild+i)+j); *(sparents+k)=i;
	}
    }
  
  // Get the parents of the gene tree.
  parents=malloc(sizeof(int)*ngenetreenodes); ngenes=0;
  for (i=0; i<=(ngenetreenodes-1); i++)
    *(parents+i)=-1;
  for (i=0; i<=(ngenetreenodes-1); i++)
    {
      if (**(genetreechild+i)==0)
	ngenes++;
      for (j=1; j<=**(genetreechild+i); j++)
	{
	  k=*(*(genetreechild+i)+j); *(parents+k)=i;
	}
    }  

  // Update the branch length.
  for (i=0; i<=(nspeciestreenodes-1); i++)
    *(speciest+i)=0;

  for (i=0; i<=(n-1); i++)
    {
      int g, pg;
      
      g=*(order+i); pg=*(parents+g);
      
      // A valid edge.
      if ((g>=0)&&(pg>=0)&&(g<ngenetreenodes)&&(pg<ngenetreenodes))
	{
	  int s, ps;
	  
	  s=*(mapping+g); ps=*(mapping+pg);	  
	  if ((s>=0)&&(s<nspeciestreenodes)&&(ps>=0)&&(ps<nspeciestreenodes))
	    {
	      // An edge in the gene tree maps to an edge in the species tree.
	      // Add the branch length of the gene tree edge to the species tree edge.
	      if (ps==*(sparents+s))
		{
		  *(speciest+s)+=*(genet+g);		  
		}

	      // An edge in the gene tree maps to a path in the species tree.
	      // Add the branch length of the gene tree edge to the last edge of the path in the species tree.
	      else if (ps!=s)
		{
		  *(speciest+s)+=*(genet+g);
		}
		  
	      // An edge in the gene tree maps to a node in the species tree.
	      // Find the first descendants of g which are not mapped to s.  Add the gene tree edge branch length to the last edges of the paths from g to the descendants.
	      else
		{		  
		  recurse_dist(ngenetreenodes,genetreechild,mapping,g,s,*(genet+g),speciest);
		}
	    }
	}      
    }

  // Release memory.
  free(parents); free(sparents); free(order);

  // Report the species branch length.
  return;
}

// Recursively trace a gene tree if the child of the current node maps to the same species node as the current node.  When stop, add the branch length of the last edge in the species tree.
void recurse_dist(int ngenetreenodes, int **genetreechild, int *mapping, int curnode, int s, double t, double *speciest)
{
  int i, j, k, s1;

  s1=*(mapping+curnode);

  // Stop when the current node is a leaf or the current node is not mapped to s.
  if ((**(genetreechild+curnode)==0)||(s1!=s))
    {
      *(speciest+s1)+=t;      		
      return;
    }

  // Otherwise continue to recurse on its children.
  else
    {
      for (i=1; i<=**(genetreechild+curnode); i++)
	{
	  j=*(*(genetreechild+curnode)+i);
	  recurse_dist(ngenetreenodes,genetreechild,mapping,j,s,t,speciest);
	}
    }
  
  return;
}

// Obtain the traverse order of a tree.
void traverseorder(int nnodes, int **child, int *order_pnt)
{
  int i, j, k, *order, length, curnode, *parents;

  parents=malloc(sizeof(int)*nnodes);
  for (i=0; i<=(nnodes-1); i++)
    *(parents+i)=-1;
  for (i=0; i<=(nnodes-1); i++)
    {
      for (j=1; j<=**(child+i); j++)
	*(parents+*(*(child+i)+j))=i;
    }

  order=malloc(sizeof(int)*nnodes); curnode=0; length=0;
  while (length<nnodes)
    {
      int *childinds, *remain;      
      childinds=malloc(sizeof(int)*(1+**(child+curnode)));
      for (i=1; i<=**(child+curnode); i++)
	*(childinds+i-1)=*(*(child+curnode)+i);

      remain=setdiff(**(child+curnode),childinds,length,order,&k);
      
      if ((**(child+curnode)==0)||(k==0))
	{
	  length++; *(order+length-1)=curnode;	  
	  if (*(parents+curnode)!=-1)
	    curnode=*(parents+curnode);
	}
      else	
	curnode=*remain;
      free(childinds); free(remain);
    }

  free(parents);
  *order_pnt=order;
  return;
}

// Obtain the partial traverse order of a tree.
// Only concern a subset of leaf nodes.
void partialtraverseorder(int nnodes, int **child, int nrep, int *rep, int *retn, int *order_pnt)
{
  int i, j, k, l, n, *order, length, curnode, *parents, **descendants, nseq, *considered;
  int *fullorder, pnt1;

  parents=malloc(sizeof(int)*nnodes);
  for (i=0; i<=(nnodes-1); i++)
    *(parents+i)=-1;
  nseq=0;
  for (i=0; i<=(nnodes-1); i++)
    {
      for (j=1; j<=**(child+i); j++)
	*(parents+*(*(child+i)+j))=i;
      if (**(child+i)==0)
	nseq++;
    }
  
  // Obtain the descendants of each node.
  // Use the travered order to find the descendants.  Faster than tracing each lineage separately.
  traverseorder(nnodes,child,&pnt1);
  fullorder=pnt1;
  descendants=malloc(sizeof(int)*nnodes);  
  for (i=0; i<=(nseq-1); i++)
    {
      *(descendants+i)=malloc(sizeof(int)*2); **(descendants+i)=1;
      *(*(descendants+i)+1)=i;
    }
  for (i=0; i<=(nnodes-1); i++)
    {
      int g=*(fullorder+i), gc;
     
      if (**(child+g)>0)
	{
	  k=0;
	  for (j=1; j<=**(child+g); j++)
	    {
	      gc=*(*(child+g)+j); k+=**(descendants+gc);
	    }
	  *(descendants+g)=malloc(sizeof(int)*(1+k));
	  **(descendants+g)=k; k=1;
	  for (j=1; j<=**(child+g); j++)
	    {
	      gc=*(*(child+g)+j); 
	      for (l=1; l<=**(descendants+gc); l++)
		{
		  *(*(descendants+g)+k)=*(*(descendants+gc)+l); k++;
		}
	    }
	}      
    }
  free(fullorder);
  
  // Only consider the nodes which are selected leaves or are ancestors of the selected leaves.
  considered=malloc(sizeof(int)*nnodes); n=0;
  for (i=0; i<=(nnodes-1); i++)
    {
      *(considered+i)=0;
      if ((**(child+i)==0)&&(inlist(nrep,rep,i)==1))
	*(considered+i)=1;
      else
	{
	  j=0;
	  while ((j<nrep)&&(*(considered+i)==0))
	    {
	      if (inlist(**(descendants+i),*(descendants+i)+1,*(rep+j))==1)
		*(considered+i)=1;
	      j++;
	    }
	}
      if (*(considered+i)==1)
	n++;
    }
 
  order=malloc(sizeof(int)*n); curnode=*(rep+0); length=0;
  while (length<n)
    {
      int *childinds, *remain;      
      childinds=malloc(sizeof(int)*(1+**(child+curnode)));
      for (i=1; i<=**(child+curnode); i++)
	*(childinds+i-1)=*(*(child+curnode)+i);

      remain=setdiff(**(child+curnode),childinds,length,order,&k);

      // Count the number of the remaining children which are considered.
      l=0;
      for (i=0; i<=(k-1); i++)
	if (*(considered+*(remain+i))==1)
	  l++;

      // The current node is a child, or all its children which are considered are already visited: trace its parent.
      if ((**(child+curnode)==0)||(l==0))
	{
	  length++; *(order+length-1)=curnode;	  
	  if (*(parents+curnode)!=-1)
	    curnode=*(parents+curnode);
	}
      // Descend to the child which is considered.
      else
	{	  
	  i=0; j=-1;
	  while ((i<k)&&(j==-1))
	    {
	      if (*(considered+*(remain+i))==1)
		j=*(remain+i);
	      i++;
	    }	  
	  curnode=j;
	}
      free(childinds); free(remain);
    }
  free(parents); free(considered);
  for (i=0; i<=(nnodes-1); i++)
    free(*(descendants+i));
  free(descendants);
  *retn=length; *order_pnt=order;
  return;
}
